{ Copyright (C) 1997-2004 John Coppens (john@jcoppens.com)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Library General Public License for more details.
  
  You should have received a copy of the GNU Library General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}  

{ unit i2c:
  builds on the ppar unis and derives a new object, i2c which included
  most of the basic funcionality of the i2c(tm) bus.
  The object includes methods for initialization and release of the
  port. At the next level, simple manipulation of the SDA and SCL
  signals, and, its highest level, it can send and receive a byte
  to the bus, with acknowledging.
  
  There are a couple of compile flags for debugging.
}
  
unit i2c;

{ -------------------------- } interface { ------------------------- }

{ $define debug}
{ $define bitlog}
{ $define en_clockhold}

{$ifdef bitlog}
  uses
    PPar, crt;
{$else}
  uses
    PPar;
{$endif}
  
const
  I2C_SDAOUT = 7;		{ Bit 7 of the data port }
  I2C_SDAIN  = PPAR_BUSY;
  I2C_SCLOUT = PPAR_SELECT;
  I2C_SCLIN  = PPAR_ERROR;
  
  I2C_READ = $01;
  I2C_WRITE = $00;
  
  { These compensate the external hardware connected to the parport }
  I2C_INV_SDAIN = true;
  I2C_INV_SDAOUT = true;
  I2C_INV_SCLOUT = true;
  I2C_INV_SCLIN = false;	{ There are two inverters, so no inversion }
  
  I2C_DEL_90 = 4700;
  I2C_DEL_91 = 4000;
  I2C_DEL_92 = 4700;
  I2C_DEL_93 = 4000;
  
  prv_sda: char = '-';
  prv_scl: char = '-';
  logfn = '/tmp/i2c_log.txt';
  
type
  i2c_port = object (par_port)
    i2c_addr: word;
    
    procedure Initialize(paddr, iaddr: word);
    procedure Release;
    
    procedure SetSDA(state: boolean);	{ Set SDA to the new state }
    function  GetSDA: boolean;		{ Get the actual SDA state }
    procedure SetSCL(state: boolean);	{ Set SCL to the new state }
    function  GetSCL: boolean;		{ Get the actual SCL state }
    procedure RaiseSCL;
        { Raise SCL and wait for it to rise (permitting the peripheral
          to 'hold' the bus }
                    
    procedure Start;			{ Generate a start condition }
    procedure Stop;			{ Generate a stop condition }

    function  ByteSent(b: byte): boolean;
        { Send a byte over the bus. Returns true or false depending if
          the peripheral confirmed reception }
    function  ByteReceived(var b: byte; do_ack: boolean): boolean;
        { Receive a byte from the bus, as confirm it with ACK or NAK,
          accoding to the parameters do_ack. Always return true (for
          the moment, until timeouts have been added for RaiseSCL }
          
    function  ReadStarted: boolean;     
        { Send an address over the bus, with bit 0 set to READ. Return
          true if it was claimed by a peripheral }
    function  WriteStarted: boolean;     
        { Send an address over the bus, with bit 0 set to WRITE. Return
          true if it was claimed by a peripheral }
          
    function  Responds: boolean;
        { Just send a WRITE addrees, and check if there's a reply.
          Immediately aborts the transfer with a Stop condition.
          NOTE: This may now work correctly with all peripherals! }

    procedure Log(state, isSDA: boolean);
        { Auxiliary routine to log a change of state. Only necessary
          for the debugging routines. }
  end;


{ -------------------------- } implementation { -------------------- }

  uses
    BaseUnix;

{ There doesn't seem to be a 'usleep' in Free Pascal's libraries, so
  I used 'nanosleep' to simulate it }

procedure usleep(time: longint);
var
  req, rem: timespec;
begin
  with req do begin
    tv_sec := 0;
    tv_nsec := time * 1000;
  end;
  fpNanoSleep(@req, @rem);
end;

{ Initialize the hardware port, and set the i2c address of this device.
  Input	paddr: parallel port address
        iaddr: i2c device address (lower 7 bits)
}

procedure i2c_port.Initialize(paddr, iaddr: word);
begin
  inherited Initialize(paddr);
  i2c_addr := iaddr;
  {$ifdef debug}
    writeln('i2cport Initialize. Port: ', HexStr(paddr, 4), '  i2c: ',
            HexStr(iaddr, 2));
  {$endif}
  SetSCL(false);
  SetSDA(true);
  SetSCL(true);
end;

procedure i2c_port.Release;
begin
  inherited Release;
end;

procedure i2c_port.SetSDA(state: boolean);
begin
  SetDataBit(I2C_SDAOUT, state xor I2C_INV_SDAOUT);
{$ifdef bitlog}
  Log(state, true);
{$endif}
end;

function i2c_port.GetSDA: boolean;
begin
  GetSDA := GetStatusBit(I2C_SDAIN) xor I2C_INV_SDAIN;
end;

procedure i2c_port.SetSCL(state: boolean);
begin
  SetControlBit(I2C_SCLOUT, state xor I2C_INV_SCLOUT);
{$ifdef bitlog}
  Log(state, false);
{$endif}
end;

function i2c_port.GetSCL: boolean;
begin
  GetSCL := GetStatusBit(I2C_SCLIN) xor I2C_INV_SCLIN;
end;

procedure i2c_port.RaiseSCL;
begin
  SetSCL(true);
  {$ifdef en_clockhold }
  repeat until GetSCL;
  {$endif }
end;  

procedure i2c_port.Start;
begin
  SetSCL(false);
  SetSDA(true);
  SetSCL(true);
  usleep(I2C_DEL_90);
  SetSDA(false);
  usleep(I2C_DEL_91);
  SetSCL(false);
end;

procedure i2c_port.Stop;
begin
  SetSCL(false);
  SetSDA(false);
  SetSCL(true);
  usleep(I2C_DEL_92);
  SetSDA(true);
  usleep(I2C_DEL_93);
end;

function i2c_port.ByteSent(b: byte): boolean;
var
  bit: integer;
begin
  {$ifdef debug}
  writeln('ByteSent: ', HexStr(b, 2));
  {$endif}
  for bit := 7 downto 0 do begin
    SetSDA((b and (1 shl bit)) <> 0);
    RaiseSCL;
    SetSCL(false);
  end;
  SetSDA(true);			{ Set true to enable reception of ACK }
  RaiseSCL;
  ByteSent := not GetSDA;
  SetSCL(false);
end;

function i2c_port.ByteReceived(var b: byte; do_ack: boolean): boolean;
var
  bit: integer;
  data: byte;
begin
  data := 0;
  {$ifdef debug}
  writeln('ByteReceived: will ACK: ', do_ack);
  {$endif}
  SetSDA(true);		{ Be sure to float the SDA line }
  for bit := 0 to 7 do begin
    RaiseSCL;
    data := (data shl 1);
    if GetSDA then
      data := data or 1;
    SetSCL(false);
  end;
  SetSDA(not do_ack);	{ Send ACK to peripheral, if more data is needed }
  RaiseSCL;
  SetSCL(false);
  SetSDA(true);
  
  b := data;
  ByteReceived := true;
  {$ifdef debug}
  writeln('ByteReceived: received ', HexStr(b, 2));
  {$endif}
end;

function  i2c_port.ReadStarted: boolean;
begin
  ReadStarted := ByteSent((i2c_addr shl 1) + I2C_READ);
end;

function  i2c_port.WriteStarted: boolean;
begin
  WriteStarted := ByteSent((i2c_addr shl 1) + I2C_WRITE);
end;

function  i2c_port.Responds: boolean;
begin
  Responds := false;
  Start;
  if ByteSent((i2c_addr shl 1) + I2C_WRITE) then
    Responds := true;
  Stop;
end;

procedure i2c_port.Log(state, isSDA: boolean);
const
  c: array [false..true] of char = ('0', '1');
var
  logf: text;
begin
{$ifdef bitlog}
  assign(logf, logfn);
  append(logf);
  if isSDA then
    prv_sda := c[state]
  else
    prv_scl := c[state];

  writeln(logf, prv_sda:2, prv_scl:4);
  close(logf);
{$endif}
end;

var
  logf: text;

begin
{$ifdef bitlog }
  assign(logf, logfn);
  rewrite(logf);
  writeln(logf, 'SDA SCL');
  close(logf);
{$endif}
end.
