{ Copyright (C) 1997-2004 John Coppens (john@jcoppens.com)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Library General Public License for more details.
  
  You should have received a copy of the GNU Library General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
} 

{ unit Ppar

  Manages the parallel port. It provides simple byte read/write
  functions, and more sophisticated bit set/reset functions for
  manipulating single bits in the Status, Control and Data registers
  of the SPP (Simple Parallel Port).
 
  The program can only be run under root priviledges in Linux, as
  the unit requests access to the parallel port hardware to the
  operating system. 
}

unit Ppar;

{ ------------------------ } interface { ------------------------------ }

{$ifdef Linux}
uses
  Ports, x86;
{$endif}

const
  PPAR_SELECT = 3;		{ output, control port }
  PPAR_ERROR = 3;		{ input, status port }
  PPAR_BUSY = 7;		{ input, status port }

  PPAR_STATUS = 1;
  PPAR_CONTROL = 2;

  PPAR_INV_STATUS = $80;
  PPAR_INV_CONTROL = $0B;

type
  par_port = object
    addr: word;
    ctl_state,
    data_state: byte;

    procedure Initialize(paddr: word);
        { Set the initial state of the parallel port, and manages the
          operating system call to request access. }
    procedure Release;
        { Returns control to the operating system }

    procedure SetData(data: byte);
        { Set the data port to 'data' }
    function GetData: byte;
        { Read the data port. }
        
    procedure SetDataBit(bitnr: integer; new_state: boolean);
        { Set data bit 'bitnr' to the new state. }
    procedure SetControlBit(bitnr: integer; new_state: boolean);
        { Set control bit 'bitnr' to the new state. Note that this functions
          takes into account any hardware inversions, and actually sets
          the output line. }

    function GetDataBit(bitnr: integer): boolean;
        { Get data bit 'bitnr'. }
    function GetControlBit(bitnr: integer): boolean;
        { Get Control bit 'bitnr'. Though it doesn't make too much sense,
          it does come in handy sometimes. Note that the function takes
          into account any inversions, and actually returns the output line
          state. }
    function GetStatusBit(bitnr: integer): boolean;
        { Get Control bit 'bitnr'. Note that the function takes
          into account any inversions, and actually returns the output line
          state. }
  end;

{ ---------------------- } implementation { --------------------------- }

procedure par_port.Initialize(paddr: word);
var
  err: integer;
begin
  addr := paddr;
  data_state := $ff;
{$ifdef fpc}
  err := fpioperm(paddr, 3, 1);
  if err <> 0 then begin
    writeln('Cannot IOPerm the parallel port');
    halt;
  end;
{$endif}
  port[addr] := data_state;
end;

procedure par_port.Release;
begin
{$ifdef fpc}
  fpioperm(addr, 3, 0);
{$endif}
end;
  

procedure par_port.SetData(data: byte);
begin
  data_state := data;
  port[addr] := data_state;
end;

function par_port.GetData: byte;
begin
  GetData := port[addr];
end;


procedure par_port.SetDataBit(bitnr: integer; new_state: boolean);
begin
  if new_state then
    data_state := data_state or (1 shl bitnr)
  else
    data_state := data_state and not(1 shl bitnr);
  port[addr] := data_state;		// No inversion for data port
end;

procedure par_port.SetControlBit(bitnr: integer; new_state: boolean);
begin
  if new_state then
    ctl_state := ctl_state or (1 shl bitnr)
  else
    ctl_state := ctl_state and not(1 shl bitnr);
  port[addr+PPAR_CONTROL] := ctl_state xor PPAR_INV_CONTROL;
end;

function par_port.GetDataBit(bitnr: integer): boolean;
begin
  GetDataBit := (data_state and (1 shl bitnr))  <> 0;
end;

function par_port.GetControlBit(bitnr: integer): boolean;
begin
  GetControlBit := (port[addr+PPAR_CONTROL] and (1 shl bitnr)) <> 0;
end;

function par_port.GetStatusBit(bitnr: integer): boolean;
begin
  GetStatusBit := ((port[addr+PPAR_STATUS] xor PPAR_INV_STATUS)
                    and (1 shl bitnr)) <> 0;
end;


begin
end.
